import numpy as _np
from .baseclass import baseclass
# import copy as _copy
# from .Twiss import Twiss as _Twiss
import warnings
from .conversions import *


class Scatter(baseclass):
    _type = 'scatter'
    _order = 1

    def __init__(self, thickness=0, radlength=1, name=None, verbose=False):
        self.name   = name
        self.verbose = verbose
        self._x      = thickness
        self._X0     = radlength

    def theta_rms(self, energy_GeV):
        x_div_X0 = self._x/self._X0
        theta = 13.6e-3/energy_GeV * _np.sqrt(x_div_X0) * (1+0.038*_np.log(x_div_X0))
        return theta
    # theta_rms = property(_get_theta_rms)

    def _get_R(self):
        warnings.warn('This scatter element does not have an R matrix: returning identity instead', UserWarning, stacklevel=3)
        return _np.identity(6)
    R = property(_get_R)

    def change_E(self, old_gamma, new_gamma):
        pass

    @property
    def ele_name(self):
        name = 'SCATTER_{}_{ind:03.0f}'.format(self.name, self.ind)
        return name

    @property
    def ele_string(self, ind, gamma):
        string = (
            '{name}\t:SCATTER , &\n'
            '\t\tXP = {xp}    , &\n'
            '\t\tYP = {yp}'
            ).format(
                name = self.elename,
                xp   = self.theta_rms(gamma2GeV(gamma)),
                yp   = self.theta_rms(gamma2GeV(gamma))
                )
        string = string + self._kwargs_str

        return string



import numpy as np
from .classes import *

__all__ = ['gaussbunch_twiss']


def gaussbunch_sigmas(sx, sxp, sy, syp, sz, sd, n_parts):
    x  = np.random.normal(loc=0, scale=sx, size=n_parts)
    xp = np.random.normal(loc=0, scale=sxp, size=n_parts)

    y  = np.random.normal(loc=0, scale=sy, size=n_parts)
    yp = np.random.normal(loc=0, scale=syp, size=n_parts)

    z  = np.random.normal(loc=0, scale=sz, size=n_parts)
    d  = np.random.normal(loc=0, scale=sd, size=n_parts)

    return Bunch(x, xp, y, yp, z, d)


def gaussbunch_twiss(Beam_x, Beam_y, sz, sd, n_parts, szd=0):
    mean = np.zeros(6)
    cov_xx = Beam_x.beta*Beam_x.emit
    cov_xxp = Beam_x.avg_xxp
    cov_xpxp = Beam_x.gamma*Beam_x.emit

    cov_yy = Beam_y.beta*Beam_y.emit
    cov_yyp = Beam_y.avg_xxp
    cov_ypyp = Beam_y.gamma*Beam_y.emit

    cov = np.array(
            [
                [cov_xx  ,  cov_xxp  ,  0       ,  0        ,  0              ,  0              ] ,
                [cov_xxp ,  cov_xpxp ,  0       ,  0        ,  0              ,  0              ] ,
                [0       ,  0        ,  cov_yy  ,  cov_yyp  ,  0              ,  0              ] ,
                [0       ,  0        ,  cov_yyp ,  cov_ypyp ,  0              ,  0              ] ,
                [0       ,  0        ,  0       ,  0        ,  np.power(sz, 2) ,  szd            ] ,
                [0       ,  0        ,  0       ,  0        ,  szd            ,  np.power(sd, 2) ]
            ]
        )

    parts = np.random.multivariate_normal(mean, cov, size=n_parts)
    x = parts[:, 0]
    xp = parts[:, 1]

    y = parts[:, 2]
    yp = parts[:, 3]

    z = parts[:, 4]
    d = parts[:, 5]

    return Bunch(x, xp, y, yp, z, d)



import numpy as _np
from .driftmat import driftmat
from .baseclass import baseclass


class Focus(baseclass):
    def __init__(self, length=0, K1=0, order=1):
        self._order = int(order)
        self._type = 'focus'
        self._length = _np.float64(length)
        self.K1 = _np.float64(K1)

    def _get_K1(self):
        return self._K1

    def _set_K1(self, val):
        self._K1 = val
    K1 = property(_get_K1, _set_K1)

    def _getR(self):
        return focusmat(L=self._length, K1=self.K1, order=self._order)
    R = property(_getR, doc='The transfer matrix R for the focus.')

    def _get_length(self):
        return self._length
    length = property(_get_length)

    def change_E(self, old_gamma, new_gamma):
        old_gamma = _np.float64(old_gamma)
        new_gamma = _np.float64(new_gamma)
        self.K1 *= old_gamma / new_gamma


def focusmat(K1=0, L=0, order=1):
    if ( K1 == 0 ):
        R = driftmat(L, order)
    else:
        rtK = _np.sqrt(_np.abs(K1))
        rtK_L = rtK * L
        sin_rtK_L = _np.sin(rtK_L)
        cos_rtK_L = _np.cos(rtK_L)
        sinh_rtK_L = _np.sinh(rtK_L)
        cosh_rtK_L = _np.cosh(rtK_L)
        R_f = _np.array(
            [[ cos_rtK_L    , sin_rtK_L/rtK ],
            [  -rtK*sin_rtK_L , cos_rtK_L   ]]
            )
        R_d = R_f
        if ( K1 > 0 ):
            R = _np.zeros([6, 6])
            R[0:2, 0:2] = R_f
            R[2:4, 2:4] = R_d
            R[4:6, 4:6] = _np.identity(2)
        elif ( K1 < 0 ):
            R = _np.zeros([6, 6])
            R[0:2, 0:2] = R_d
            R[2:4, 2:4] = R_f
            R[4:6, 4:6] = _np.identity(2)
        else:
            print('Uhm who said what now??')
            print(K1)

    return R



import numpy as _np

class BeamParams(object):
    def __repr__(self):
        return '<{} at {}; beta={}, alpha={}>'.format(self.__class__.__module__,hex(id(self)),self.beta,self.alpha)

    def __init__(self,
            beta=None,alpha=None,emit=None,emit_n=None,gamma=None,
            spotsize=None,divergence=None,r=None,avg_xxp=None):
        # If not enough info for Twiss
        emit_norm_insufficient = (emit_n==None or gamma==None)
        emit_geom_insufficient = emit==None
        emit_insufficient = (emit_geom_insufficient and emit_norm_insufficient)
        shape_insufficient = (beta==None or alpha==None)
        twiss_insufficient = (shape_insufficient or emit_insufficient)
        moments_insufficient = (spotsize==None or divergence==None or (r==None and avg_xxp==None))

        # Validation that there's enough info to construct everything
        if twiss_insufficient and moments_insufficient:
            raise ValueError('Not enough information to construct Twiss values')

        if twiss_insufficient:
            if avg_xxp!=None:
                emit = _np.sqrt(spotsize**2 * divergence**2 - avg_xxp**2)
                corr_sign = _np.sign(avg_xxp)
            elif r!=None:
                emit = spotsize * divergence * _np.sqrt(1-r**2)
                corr_sign = _np.sign(r)
            else:
                raise ValueError('Missing avg_xxp/r term')
            beta = spotsize**2/emit
            gamma = divergence**2/emit
            alpha = -corr_sign * _np.sqrt(gamma*beta-1)
        else:
            def _get_geom_emit(emit=None,emit_n=None,gamma=None):
                if emit==None and emit_n==None:
                    raise ValueError('Did not specify an emittance!')
                if emit!=None and emit_n!=None:
                    raise ValueError('Specified too many emittance values!')
                if emit_n!=None:
                    # print emit_n
                    emit = _np.float64(emit_n)/_np.float64(gamma)
                else:
                    emit = _np.float64(emit)
                return emit
            emit=_get_geom_emit(emit=emit,emit_n=emit_n,gamma=gamma)

        self.beta=_np.float64(beta)
        self.alpha=_np.float64(alpha)
        self.emit=_np.float64(emit)

    def set_emit_n(self,emit_n,gamma):
        self.emit=_np.float64(emit_n/gamma)


    # Definte beta property
    # Validate beta > 0
    def _get_beta(self):
        return self._beta
    def _set_beta(self,value):
        if not (value > 0):
            raise ValueError('Beta must be greater than zero: requested beta={}.'.format(value))
        self._beta=value
    beta = property(_get_beta,_set_beta)
    
    def _get_betastar(self):
        return 1.0/self.gamma
    betastar=property(_get_betastar)

    def _get_sstar(self):
        return self.alpha/self.gamma
    sstar=property(_get_sstar)

    # Define gamma property
    # Derived from beta and alpha
    def _get_gamma(self):
        return (1+_np.power(self.alpha,2))/self.beta
    gamma=property(_get_gamma)

    # Define T matrix
    # Derived from beta, alpha, and gamma
    def _get_T(self):
        return _np.array([[self.beta,-self.alpha],[-self.alpha,self.gamma]])
    T=property(_get_T)

    # Define transport method
    # Returns new twiss given R matrix.
    def transport(self,R):
        T2 = _np.dot(_np.dot(R,self.T),_np.transpose(R))
        betaf = T2[0,0]
        alphaf = -T2[0,1]
        return BeamParams(beta=betaf,alpha=alphaf,emit=self.emit)

    # Define spotsize method
    def _get_spotsize(self):
        return _np.sqrt(self.beta*self.emit)
    spotsize=property(_get_spotsize)

    def _get_divergence(self):
        return _np.sqrt(self.gamma*self.emit)
    def _set_divergence(self,value):
        self.emit = value**2/self.gamma
        # print self.divergence
        # print self.emit
    divergence=property(_get_divergence,_set_divergence)

    def _get_avg_xxp(self):
         return -self.alpha*self.emit
    avg_xxp = property(_get_avg_xxp)

    def _get_minspotsize(self):
        return _np.sqrt(self.betastar*self.emit)
    minspotsize = property(_get_minspotsize)



import ipdb
import jinja2 as jj
import logging
import os
import pkg_resources
import shlex
import shutil
import tempfile
import subprocess

logger=logging.getLogger(__name__)
loggerlevel = logging.DEBUG
#  loggerlevel = 9

__all__ = ['elegant_sim']

def elegant_sim(beamline,dir=None,filename=None,**kwargs):
    # ======================================
    # If path isn't specified, create
    # temporary one
    # ======================================
    if dir is None:
        path = tempfile.mkdtemp()
        logger.log(level=loggerlevel,msg='Path is: {}'.format(path))
    else:
        path = dir

    # ======================================
    # Open up a temporary file in the path
    # ======================================
    if filename is None:
        fd,filename = tempfile.mkstemp(dir=path,prefix='out_',suffix='.ele')
        f = os.fdopen(fd,'w+')
    else:
        f = open(os.path.join(path,filename),'w+')
    logger.log(level=loggerlevel,msg='Filename is: {}'.format(filename))

    basename = os.path.basename(filename)
    root,ext = os.path.splitext(basename)
    lte_name = os.path.join(path,'{}.lte'.format(root))
    beamline.elegant_lte(filename=lte_name)
    
    # ======================================
    # Get template path from package and
    # load template
    # ======================================
    templates_path = pkg_resources.resource_filename(__name__,'templates')
    loader = jj.FileSystemLoader(templates_path)
    env = jj.Environment(loader=loader)
    template = env.get_template('template.ele')
    
    # ======================================
    # Generate ele file
    # ======================================
    template.stream(filename=lte_name,matched=0,**kwargs).dump(f)
    f.close()

    # ======================================
    # Run Elegant
    # ======================================
    command = 'elegant {}'.format(filename)
    logger.log(level=loggerlevel,msg='Command is: {}'.format(command))

    log_name = os.path.join(path,'{}.log'.format(root))
    f_log = open(log_name,'w+')
    logger.log(level=loggerlevel,msg='Log file is: {}'.format(log_name))

    #  fnull = open(os.devnull,'w')
    cwdu = os.getcwdu()
    os.chdir(path)
    subprocess.call(shlex.split(command),stdout=f_log,stderr=f_log)
    os.chdir(cwdu)

    # ======================================
    # Clean temporary directory if necessary
    # ======================================
    if dir is None:
        shutil.rmtree(path)

    return path,root,ext



import numpy as _np
import warnings

class Twiss(object):
    def __repr__(self):
        return '<{} at {}; beta={}, alpha={}>'.format(self.__class__.__module__,hex(id(self)),self.beta,self.alpha)

    def __init__(self,beta,alpha):
        warnings.warn('Twiss is deprecated, use BeamParams instead',DeprecationWarning)
        self.beta=_np.float64(beta)
        self.alpha=_np.float64(alpha)

    # Definte beta property
    # Validate beta > 0
    def _set_beta(self,value):
        if not (value > 0):
            raise ValueError('Beta must be greater than zero: requested beta={}.'.format(value))
        self._beta=value
    def _get_beta(self):
        return self._beta
    beta = property(_get_beta,_set_beta)
    
    def _get_betastar(self):
        return 1/self.gamma
    betastar=property(_get_betastar)

    def _get_sstar(self):
        return self.alpha/self.gamma
    sstar=property(_get_sstar)

    # Define gamma property
    # Derived from beta and alpha
    def _get_gamma(self):
        return (1+_np.power(self.alpha,2))/self.beta
    gamma=property(_get_gamma)

    # Define T matrix
    # Derived from beta, alpha, and gamma
    def _get_T(self):
        return _np.array([[self.beta,-self.alpha],[-self.alpha,self.gamma]])
    T=property(_get_T)

    # Define transport method
    # Returns new twiss given R matrix.
    def transport(self,R):
        T2 = _np.dot(_np.dot(R,self.T),_np.transpose(R))
        betaf = T2[0,0]
        alphaf = -T2[0,1]
        return Twiss(betaf,alphaf)

    # Define spotsize method
    # Returns spot size given an emittance
    def spotsize(self,emit=None,emit_n=None,gamma=None):
        return _np.sqrt(self.beta*self._get_geom_emit(emit=emit,emit_n=emit_n,gamma=gamma))

    def _get_geom_emit(self,emit=None,emit_n=None,gamma=None):
        if emit==None and emit_n==None:
            raise ValueError('Did not specify an emittance!')
        if emit!=None and emit_n!=None:
            raise ValueError('Specified too many emittance values!')
        if emit_n!=None:
            # print emit_n
            emit = _np.float64(emit_n)/_np.float64(gamma)
        else:
            emit = _np.float64(emit)
        return emit

    def minspotsize(self,emit=None,emit_n=None,gamma=None):
        return _np.sqrt(self.betastar*self._get_geom_emit(emit=emit,emit_n=emit_n,gamma=gamma))



from .driftmat import *
from .bend import *
from .quadmat import *
from .focusmat import *
from .beamline import *
from .scatter import *
# from runsim import runsim
from .Twiss import *
from .BeamParams import *
from .conversions import *
from .elegant_sim import *
from .bunchgen import *
from .classes import *



import numpy as _np
from .consts import *


def gamma2GeV(gamma):
    n_pts = _np.size(gamma)
    if n_pts > 1:
        for i, val in enumerate(gamma):
            gamma[i] = gamma2GeV(val)
        return gamma
    else:
        gamma = _np.float(gamma)
        return gamma*electron_mc2_gev


def GeV2gamma(E):
    if _np.size(E) == 1:
        E = _np.float(E)
    else:
        E = _np.array(E, dtype='float')
    return E/electron_mc2_gev


def GeV2joule(E):
    return eV2joule(E) * 1e9


def eV2joule(E):
    if _np.size(E) == 1:
        E = _np.float(E)
    else:
        E = _np.array(E, dtype='float')

    return - E * e



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

__all__ = ['Bunch']

class Bunch(object):
    def __init__(self,x,xp,y,yp,z,delta,parts=None):
        self._parts = np.zeros(6,np.size(x))

    def _get_x(self):
        return self._parts[0,:]
    def _set_x(self,val):
        self._parts[0,:] = val
    x = property(_get_x,_set_x)

    def hist_xy(self,ax=None,fig=None,bins=50):
        if ax is None:
            if fig is None:
                ax = plt.gca()
            else:
                gs = gridspec.GridSpec(1,1)
                ax = fig.add_subplot(gs[0,0])

        hst = ax.hist2d(self.x,self.y,bins=bins)

        return hst

    def prop(self,beamline,gamma=None):
        if gamma is None:
            gamma = beamline.gamma



class baseclass(object):
    def __init__(self,length=0,order=1,name=None):
        self.ind     = None
        self.name    = name
        self._length = length
        self._order  = order
        self._type   = None
        self._ele_type = None

    def _not_allowed(self,*args,**kwargs):
        raise AttributeError('This attribute is not available for element type: {}'.format(self._type))

    K1        = property(_not_allowed,_not_allowed)
    angle     = property(_not_allowed,_not_allowed)
    rotate    = property(_not_allowed,_not_allowed)
    thickness = property(_not_allowed,_not_allowed)
    radlength = property(_not_allowed,_not_allowed)

    @property
    def _kwargs_str(self):
        string = ''
        for kname,kvalue in self._kwargs.items():
            string = string + ', {} = {}'.format(kname,kvalue)

        return string
